package org.xesstandard.extensions;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import org.xesstandard.model.XesAttribute;
import org.xesstandard.model.XesComponent;
import org.xesstandard.model.XesEvent;
import org.xesstandard.model.XesLog;
import org.xesstandard.model.attributes.XesStringAttribute;
import org.xesstandard.model.attributes.impl.XesStringAttributeImpl;
import org.xesstandard.model.impl.XesExtensionImpl;

public class XesLifecycleExtension extends XesExtensionImpl {

	public final static String MODEL = "lifecycle:model";
	public final static String TRANSITION = "lifecycle:transition";
	public final static String STATE = "lifecycle:state";

	private final static XesLifecycleExtension singleton = new XesLifecycleExtension();

	private Map<String, XesLifecycleModel> models;
	private Map<String, XesLifecycleState> states;
	private Map<String, XesLifecycleTransition> transitions;

	private XesLifecycleExtension() {
		super(true);
		setName("Lifecycle");
		setPrefix("lifecycle");
		try {
			setURI(new URI("http://www.xes-standard.org/lifecycle.xesext"));
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}

		models = new HashMap<String, XesLifecycleModel>();
		for (XesLifecycleModel model : XesLifecycleModel.values()) {
			models.put(model.getModel(), model);
		}

		states = new HashMap<String, XesLifecycleState>();
		for (XesLifecycleState state : XesLifecycleState.values()) {
			states.put(state.getState(), state);
		}

		transitions = new HashMap<String, XesLifecycleTransition>();
		for (XesLifecycleTransition transition : XesLifecycleTransition.values()) {
			transitions.put(transition.getTransition(), transition);
		}

		XesAttribute attribute = new XesStringAttributeImpl();
		attribute.setKey(MODEL);
		addLogAttribute(attribute);
		addAlias(attribute, "EN", "Lifecycle Model");
		addAlias(attribute, "DE", "Lebenszyklus-Modell");
		addAlias(attribute, "FR", "Mod\u00e8le du Cycle Vital");
		addAlias(attribute, "ES", "Modelo de Ciclo de Vida");
		addAlias(attribute, "PT", "Modelo do Ciclo de Vida");

		attribute = new XesStringAttributeImpl();
		attribute.setKey(TRANSITION);
		addEventAttribute(attribute);
		addAlias(attribute, "EN", "Lifecycle Transition");
		addAlias(attribute, "DE", "Lebenszyklus-Transition");
		addAlias(attribute, "FR", "Transition en Cycle Vital");
		addAlias(attribute, "ES", "Transici\u00f3n en Ciclo de Vida");
		addAlias(attribute, "PT", "Transi\u00e7\u00e3o do Ciclo de Vida");

		attribute = new XesStringAttributeImpl();
		attribute.setKey(STATE);
		addEventAttribute(attribute);
		addAlias(attribute, "EN", "Lifecycle State");
	}

	/**
	 * @return The single instance of this extension.
	 */
	public static XesLifecycleExtension instance() {
		return singleton;
	}

	private XesStringAttribute setTransition(XesComponent component, String key, XesLifecycleTransition value) {
		XesStringAttribute attribute = new XesStringAttributeImpl();
		attribute.setKey(key);
		attribute.setValue(value.getTransition());
		component.addAttribute(attribute);
		return attribute;
	}
	
	private XesStringAttribute setState(XesComponent component, String key, XesLifecycleState value) {
		XesStringAttribute attribute = new XesStringAttributeImpl();
		attribute.setKey(key);
		attribute.setValue(value.getState());
		component.addAttribute(attribute);
		return attribute;
	}
	
	private XesStringAttribute setModel(XesComponent component, String key, XesLifecycleModel value) {
		XesStringAttribute attribute = new XesStringAttributeImpl();
		attribute.setKey(key);
		attribute.setValue(value.getModel());
		component.addAttribute(attribute);
		return attribute;
	}
	
	/**
	 * Returns the lifecycle model for the given log, that is, the value of the
	 * attribute with key "lifecycle:model".
	 * 
	 * @param log
	 *            The given log.
	 * @return The model for the given log.
	 */
	public XesLifecycleModel getModel(XesLog log) {
		String value = log.getString(MODEL);
		if (value != null) {
			return models.get(value);
		}
		return null;
	}

	/**
	 * Sets the lifecycle model for the given log, that is, the value of the
	 * attribute with key "lifecycle:model".
	 * 
	 * @param log
	 *            The given log.
	 * @param value
	 *            The model to set.
	 * @return The attribute containing the model.
	 */
	public XesStringAttribute setModel(XesLog log, XesLifecycleModel model) {
		return setModel(log, MODEL, model);
	}

	/**
	 * Returns the transition for the given event, that is, the value of the
	 * attribute with key "lifecycle:transition". Assumes that the "standard"
	 * lifecycle model is used for the log.
	 * 
	 * @param event
	 *            The given event.
	 * @return The transition for the given event.
	 */
	public XesLifecycleTransition getTransition(XesEvent event) {
		String value = event.getString(TRANSITION);
		if (value != null) {
			return transitions.get(value);
		}
		return null;
	}

	/**
	 * Sets the transition for the given event, that is, the value of the
	 * attribute with key "lifecycle:model". Assumes that the "standard"
	 * lifecycle model is used for the log.
	 * 
	 * @param event
	 *            The given event.
	 * @param transition
	 *            The transition to set.
	 * @return The attribute containing the transition.
	 */
	public XesStringAttribute setTransition(XesEvent event, XesLifecycleTransition transition) {
		return setTransition(event, TRANSITION, transition);
	}

	/**
	 * Returns the state for the given event, that is, the value of the
	 * attribute with key "lifecycle:state". Assumes that the "bpaf" lifecycle
	 * model is used for the log.
	 * 
	 * @param event
	 *            The given event.
	 * @return The state for the given event.
	 */
	public XesLifecycleState getState(XesEvent event) {
		String value = event.getString(STATE);
		if (value != null) {
			return states.get(value);
		}
		return null;
	}

	/**
	 * Sets the state for the given event, that is, the value of the attribute
	 * with key "lifecycle:model". Assumes that the "bpaf" lifecycle model is
	 * used for the log.
	 * 
	 * @param event
	 *            The given event.
	 * @param state
	 *            The state to set.
	 * @return The attribute containing the state.
	 */
	public XesStringAttribute setState(XesEvent event, XesLifecycleState state) {
		return setState(event, STATE, state);
	}
}
