package org.xesstandard.extensions;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.xesstandard.model.XesAttribute;
import org.xesstandard.model.XesComponent;
import org.xesstandard.model.XesEvent;
import org.xesstandard.model.XesTrace;
import org.xesstandard.model.attributes.XesIdAttribute;
import org.xesstandard.model.attributes.XesIntegerNumberAttribute;
import org.xesstandard.model.attributes.impl.XesIdAttributeImpl;
import org.xesstandard.model.attributes.impl.XesIntegerNumberAttributeImpl;
import org.xesstandard.model.impl.XesExtensionImpl;

public class XesMicroExtension extends XesExtensionImpl {

	public final static String LEVEL = "micro:level";
	public final static String PARENTID = "micro:parentId";
	public final static String LENGTH = "micro:length";

	private final static XesMicroExtension singleton = new XesMicroExtension();

	private XesMicroExtension() {
		super(true);
		setName("Micro");
		setPrefix("micro");
		try {
			setURI(new URI("http://www.xes-standard.org/micro.xesext"));
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}

		XesAttribute attribute = new XesIntegerNumberAttributeImpl();
		attribute.setKey(LEVEL);
		addEventAttribute(attribute);
		addAlias(attribute, "EN", "Event level");

		attribute = new XesIdAttributeImpl();
		attribute.setKey(PARENTID);
		;
		addEventAttribute(attribute);
		addAlias(attribute, "EN", "Id of parent event");

		attribute = new XesIntegerNumberAttributeImpl();
		attribute.setKey(LENGTH);
		addEventAttribute(attribute);
		addAlias(attribute, "EN", "Number of child events");
	}

	/**
	 * @return The single instance of this extension.
	 */
	public static XesMicroExtension instance() {
		return singleton;
	}

	private XesIntegerNumberAttribute setIntegerNumber(XesComponent component, String key, Long value) {
		XesIntegerNumberAttribute attribute = new XesIntegerNumberAttributeImpl();
		attribute.setKey(key);
		attribute.setValue(value);
		component.addAttribute(attribute);
		return attribute;
	}
	
	private XesIdAttribute setUUID(XesComponent component, String key, UUID value) {
		XesIdAttribute attribute = new XesIdAttributeImpl();
		attribute.setKey(key);
		attribute.setValue(value);
		component.addAttribute(attribute);
		return attribute;
	}
	
	/**
	 * Returns the level of the given event, that is, the value of the attribute
	 * with key "micro:level".
	 * 
	 * @param event
	 *            The given event.
	 * @return The level of the given event.
	 */
	public Long getLevel(XesEvent event) {
		Long level = event.getIntegerNumber(LEVEL);
		if (level == null) {
			level = 1L;
		}
		return level;
	}

	/**
	 * Sets the level for the given event, that is, the value of the attribute
	 * with key "micro:level".
	 * 
	 * @param event
	 *            The given event.
	 * @param value
	 *            The level to set. Must be a positive value.
	 * @return The attribute containing the level.
	 */
	public XesIntegerNumberAttribute setLevel(XesEvent event, Long value) {
		if (value > 0) {
			return setIntegerNumber(event, LEVEL, value);
		}
		return null;
	}

	/**
	 * Returns the parent id of the given event, that is, the value of the
	 * attribute with key "micro:parentId".
	 * 
	 * @param event
	 *            The given event.
	 * @return The parent id of the given event.
	 */
	public UUID getParentId(XesEvent event) {
		return event.getId(PARENTID);
	}

	/**
	 * Sets the parent id for the given event, that is, the value of the
	 * attribute with key "micro:parentId".
	 * 
	 * @param event
	 *            The given event.
	 * @param value
	 *            The parent id to set.
	 * @return The attribute containing the parent id.
	 */
	public XesIdAttribute setParentId(XesEvent event, UUID value) {
		return setUUID(event, PARENTID, value);
	}

	/**
	 * Returns the length of the given event, that is, the value of the
	 * attribute with key "micro:length".
	 * 
	 * @param event
	 *            The given event.
	 * @return The length of the given event.
	 */
	public Long getLength(XesEvent event) {
		return event.getIntegerNumber(LENGTH);
	}

	/**
	 * Sets the length for the given event, that is, the value of the attribute
	 * with key "micro:length".
	 * 
	 * @param event
	 *            The given event.
	 * @param value
	 *            The length to set.
	 * @return The attribute containing the length.
	 */
	public XesIntegerNumberAttribute setLength(XesEvent event, Long value) {
		return setIntegerNumber(event, LENGTH, value);
	}

	/**
	 * Derives the levels and lengths for the events in the given trace from
	 * their parent ids.
	 * 
	 * @param trace The given trace
	 */
	public void setLevelsAndLengths(XesTrace trace) {
		// Maps id to parent id
		Map<UUID, UUID> parentMap = new HashMap<UUID, UUID>();
		// Maps id to number of child ids
		Map<UUID, Long> lengthMap = new HashMap<UUID, Long>();
		
		// Compute both maps
		for (XesEvent event : trace.getEvents()) {
			UUID id = XesIdExtension.instance().getId(event);
			UUID parentId = getParentId(event);
			if (id != null && parentId != null) {
				// Add to parent map.
				parentMap.put(id, parentId);
				// Increase in length map, or set to 1 if it does not exist yet.
				Long length = lengthMap.get(parentId);
				if (length == null) {
					// Does not exist yet.
					length = 0L;
				}
				lengthMap.put(parentId, length + 1);
			}
		}
		// Set the maximum level an event can have.
		// This is to prevent the getLevel() method used to continue forever in case
		// there is a cycle in the parentMap.
		long maxLevel = parentMap.keySet().size() + 1;
		
		// Set the attributes
		for (XesEvent event : trace.getEvents()) {
			UUID id = XesIdExtension.instance().getId(event);
			if (id != null) {
				// Set the level
				setLevel(event, getLevel(id, parentMap, 1, maxLevel));
				// Set the length
				Long length = lengthMap.get(id);
				if (length == null) {
					length = 0L;
				}
				setLength(event, length);
			}
		}
	}

	/*
	 * Returns the level of the given id in the given map plug the given level, but
	 * topped with maxLevel. 
	 */
	private long getLevel(UUID id, Map<UUID, UUID> parentMap, long level, long maxLevel) {
		// Cycle prevention
		if (level == maxLevel) {
			return maxLevel;
		}
		UUID parentId = parentMap.get(id);
		if (parentId == null) {
			// No parent found, current level will do.
			return level;
		}
		// Level is one higher than level of parent.
		return getLevel(parentId, parentMap, level + 1, maxLevel);
	}

	/**
	 * Return the list of child events of the given event in the given trace.
	 * @param event The given event.
	 * @param trace The given trace.
	 * @return The list of child events of the given event in the given trace.
	 */
	public List<XesEvent> getChildEvents(XesEvent event, XesTrace trace) {
		List<XesEvent> childEvents = new ArrayList<XesEvent>();
		UUID parentId = XesIdExtension.instance().getId(event);
		if (parentId != null) {
			for (XesEvent childEvent : trace.getEvents()) {
				UUID id = getParentId(childEvent);
				if (id != null && id.equals(parentId)) {
					childEvents.add(childEvent);
				}
			}
		}
		return childEvents;
	}
}
