package org.xesstandard.model;

import java.util.Date;
import java.util.List;
import java.util.UUID;

/**
 * XES Component, that is, logs, traces, events, and attributes. See Clause 4.1 and Clause 4.2.
 * 
 * @author H.M.W. Verbeek
 *
 */
public interface XesComponent {

	/**
	 * Adds an attribute to the component.
	 * @param attribute The attribute to add.
	 */
	public void addAttribute(XesAttribute attribute);
	/**
	 * @return The list of attributes set for (or added to) the component.
	 * The order in the list corresponds to the order they were added.
	 */
	public List<XesAttribute> getAttributes();
	/**
	 * Returns the value of an attribute as a String.
	 * @param key The key of the attribute.
	 * @return The value (as a String) of the attribute with the given key. Null if if does not exist.
	 */
	public String getStringValue(String key);
	
	/**
	 * Returns the type of the attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The type of the attribute with the given key. Null if it does not exist.
	 */
	public XesAttributeType getType(String key);
	/**
	 * Returns the value of the String attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the String attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a String attribute.
	 */
	public String getString(String key);	
	/**
	 * Returns the value of the UTC time attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the UTC time attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a UTC time attribute.
	 */
	public Date getDateTime(String key);
	/**
	 * Returns the value of the UUID attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the UUID attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a UUID attribute.
	 */
	public UUID getId(String key);
	/**
	 * Returns the value of the Integer Number attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the Integer Number attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not an Integer Number attribute.
	 */
	public Long getIntegerNumber(String key);
	/**
	 * Returns the value of the Real Number attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the Real Number attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a Real Number attribute.
	 */
	public Double getRealNumber(String key);
	/**
	 * Returns the value of the Boolean attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The value of the Boolean attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a Booelan attribute.
	 */
	public Boolean getBoolean(String key);
	/**
	 * Returns the value attributes of the List attribute with the given key.
	 * @param key The key of the attribute.
	 * @return The list of value attributes of the List attribute with the given key. Null if it does not exist
	 * or if the attribute exists but is not a List attribute.
	 */
	public List<XesAttribute> getList(String key);
	
	/**
	 * Sets the value of a String attribute with the given key. Does not set the value if the attribute
	 * is not a String attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the String attribute. Null if the attribute does not exist, or if the
	 * attribute is not a String attribute.
	 */
	public String setString(String key, String value);
	/**
	 * Sets the value of a UTC time attribute with the given key. Does not set the value if the attribute
	 * is not a UTC time attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the UTC time attribute. Null if the attribute does not exist, or if the
	 * attribute is not a UTC time attribute.
	 */
	public Date setDateTime(String key, Date value);
	/**
	 * Sets the value of a UUID attribute with the given key. Does not set the value if the attribute
	 * is not a UUID attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the UUID attribute. Null if the attribute does not exist, or if the
	 * attribute is not a UUID attribute.
	 */
	public UUID setId(String key, UUID value);
	/**
	 * Sets the value of an Integer Number attribute with the given key. Does not set the value if the attribute
	 * is not an Integer Number attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the Integer Number attribute. Null if the attribute does not exist, or if the
	 * attribute is not an Integer Number attribute.
	 */
	public Long setIntegerNumber(String key, Long value);
	/**
	 * Sets the value of a Real Number attribute with the given key. Does not set the value if the attribute
	 * is not a Real Number attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the Real Number attribute. Null if the attribute does not exist, or if the
	 * attribute is not a Real Number attribute.
	 */
	public Double setRealNumber(String key, Double value);
	/**
	 * Sets the value of a Boolean attribute with the given key. Does not set the value if the attribute
	 * is not a Boolean attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old value of the Boolean attribute. Null if the attribute does not exist, or if the
	 * attribute is not a Boolean attribute.
	 */
	public Boolean setBoolean(String key, Boolean value);
	/**
	 * Sets the values of a List attribute with the given key. Does not set the values if the attribute
	 * is not a List attribute.
	 * @param key The given key.
	 * @param value The value to set.
	 * @return The old values of the List attribute. Null if the attribute does not exist, or if the
	 * attribute is not a List attribute.
	 */
	public List<XesAttribute> setList(String key, List<XesAttribute> values);
}
