package org.xesstandard.xml;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import org.xesstandard.factory.XesFactory;
import org.xesstandard.factory.impl.XesFactoryImpl;
import org.xesstandard.model.XesAttribute;
import org.xesstandard.model.XesExtension;
import org.xesstandard.xml.issues.XesextXmlIssueAttributeDiffers;
import org.xesstandard.xml.issues.XesextXmlIssueAttributeRequired;
import org.xesstandard.xml.issues.XesextXmlIssueLanguageCode;
import org.xesstandard.xml.issues.XesextXmlIssueOutOfPlace;
import org.xesstandard.xml.issues.XesextXmlIssueParent;
import org.xesstandard.xml.issues.XesextXmlIssueStrictlyConformingAttribute;
import org.xesstandard.xml.issues.XesextXmlIssueStrictlyConformingElement;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class XesextXmlHandler extends DefaultHandler {

	private List<XesXmlIssue> issues;
	private XesFactory factory;
	private XesExtension extension;
	private List<XesAttribute> attributeList;
	private XesAttribute currentAttribute;

	// Get all known ISO language codes.
	private static final Set<String> ISO_LANGUAGES = new HashSet<String>(Arrays.asList(Locale.getISOLanguages()));
	
	public XesextXmlHandler(XesExtension extension, List<XesXmlIssue> issues) {
		this.issues = issues;
		factory = null;
		this.extension = extension;
		attributeList = null;
		currentAttribute = null;
	}

	/*
	 * Locator to be able to store line and column numbers while parsing.
	 */
	private Locator locator;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#setDocumentLocator(org.xml.sax.
	 * Locator)
	 */
	public void setDocumentLocator(Locator locator) {
		this.locator = locator;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String,
	 * java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 */
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {

		String tagName = localName;
		if (tagName.equalsIgnoreCase("")) {
			tagName = qName;
		}

		if (tagName.equalsIgnoreCase(XesXmlTags.XESEXT)) {
			startXesextensionElement(attributes);
			factory = new XesFactoryImpl();

		} else if (tagName.equalsIgnoreCase(XesXmlTags.LOG)) {
			startLogElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.TRACE)) {
			startTraceElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.EVENT)) {
			startEventElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.META)) {
			startMetaElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.STRING)) {
			startStringElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.DATE)) {
			startDateElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.INT)) {
			startIntegerNumberElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.FLOAT)) {
			startRealNumberElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.BOOLEAN)) {
			startBooleanElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.ID)) {
			startIdElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.LIST)) {
			startListElement(attributes);

		} else if (tagName.equalsIgnoreCase(XesXmlTags.ALIAS)) {
			startAliasElement(attributes);

		} else {
			XesXmlIssue issue = new XesextXmlIssueStrictlyConformingElement(extension, tagName);
			issue.setLocation(locator);
			issues.add(issue);
		}
	}

	/*
	 * Start a xesextension element.
	 */
	private void startXesextensionElement(Attributes attributes) {
		
		// Check required name attribute. 
		String nameValue = attributes.getValue(XesXmlTags.NAME);
		if (nameValue == null) {
			// No name attribute. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, XesXmlTags.XESEXT, XesXmlTags.NAME,
					"Clause 6.1.2");
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!nameValue.equals(extension.getName())) {
			// Name differs from name in the declared extension. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueAttributeDiffers(extension, XesXmlTags.XESEXT, XesXmlTags.NAME,
					"Clause 6.1.2");
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Check required prefix attribute. 
		String prefixValue = attributes.getValue(XesXmlTags.PREFIX);
		if (prefixValue == null) {
			// No prefix attribute. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, XesXmlTags.XESEXT, XesXmlTags.PREFIX,
					"Clause 6.1.2");
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!prefixValue.equals(extension.getPrefix())) {
			// Prefix differs from prefix in the declared extension. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueAttributeDiffers(extension, XesXmlTags.XESEXT, XesXmlTags.PREFIX,
					"Clause 6.1.2");
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Check required uri attribute.
		URI uriValue = null;
		try {
			uriValue = new URI(attributes.getValue(XesXmlTags.URI));
			if (!uriValue.equals(extension.getURI())) {
				// URI differs from URI in the declared extension. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, XesXmlTags.XESEXT, XesXmlTags.URI,
						"Clause 6.1.2");
				issue.setLocation(locator);
				issues.add(issue);
			}
		} catch (URISyntaxException e) {
			// No URI, or no valid URI. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueAttributeDiffers(extension, XesXmlTags.XESEXT, XesXmlTags.URI,
					"Clause 6.1.2");
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Check for additional attributes.
		for (int i = 0; i < attributes.getLength(); i++) {
			String value = attributes.getQName(i);
			if (value.equals(XesXmlTags.NAME) || value.equals(XesXmlTags.PREFIX) || value.equals(XesXmlTags.URI)) {
			} else {
				// Found additional attribute. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value, "Clause 6.1.2");
				issue.setLocation(locator);
				issues.add(issue);
			}
		}

	}

	/*
	 * Start log element.
	 */
	private void startLogElement(Attributes attributes) {
		if (factory == null || attributeList != null) {
			// Either not in xesextension element or in other log, trace, event, or meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, XesXmlTags.LOG, XesXmlTags.XESEXT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getLogAttributes().isEmpty()) {
			// Log element may not succeed another log element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.LOG, XesXmlTags.LOG);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getTraceAttributes().isEmpty()) {
			// Log element may not succeed a trace element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.LOG, XesXmlTags.TRACE);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getEventAttributes().isEmpty()) {
			// Log element may not succeed an event element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.LOG, XesXmlTags.EVENT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getMetaAttributes().isEmpty()) {
			// Log element may not succeed a meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.LOG, XesXmlTags.META);
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Check for additional attributes.
		for (int i = 0; i < attributes.getLength(); i++) {
			String value = attributes.getQName(i);
			// Found additional attribute. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value, "Clause 6.2.2");
			issue.setLocation(locator);
			issues.add(issue);
		}

		// Set the list for the child attributes.
		attributeList = extension.getLogAttributes();
	}

	/*
	 * Start a trace element.
	 */
	private void startTraceElement(Attributes attributes) {
		if (factory == null || attributeList != null) {
			// Either not in xesextension element or in log, other trace, event, or meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, XesXmlTags.TRACE, XesXmlTags.XESEXT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getTraceAttributes().isEmpty()) {
			// Trace element may not succeed another trace element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.TRACE, XesXmlTags.TRACE);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getEventAttributes().isEmpty()) {
			// Trace element may not succeed an event element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.TRACE, XesXmlTags.EVENT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getMetaAttributes().isEmpty()) {
			// Trace element may not succeed a meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.TRACE, XesXmlTags.META);
			issue.setLocation(locator);
			issues.add(issue);
		}
		
		// Check for additional attributes.
		for (int i = 0; i < attributes.getLength(); i++) {
			String value = attributes.getQName(i);
			// Found additional attribute. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value, "Clause 6.3.2");
			issue.setLocation(locator);
			issues.add(issue);
		}
		
		// Set the list for the child attributes.
		attributeList = extension.getTraceAttributes();
	}
	
	/*
	 * Start an event element.
	 */
	private void startEventElement(Attributes attributes) {
		if (factory == null || attributeList != null) {
			// Either not in xesextension element or in log, other trace, event, or meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, XesXmlTags.EVENT, XesXmlTags.XESEXT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getEventAttributes().isEmpty()) {
			// Event element may not succeed another event element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.EVENT, XesXmlTags.EVENT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getMetaAttributes().isEmpty()) {
			// Event element may not succeed a meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.EVENT, XesXmlTags.META);
			issue.setLocation(locator);
			issues.add(issue);
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			String value = attributes.getQName(i);
			XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value, "Clause 6.4.2");
			issue.setLocation(locator);
			issues.add(issue);
		}
		
		// Set the list for the child attributes.
		attributeList = extension.getEventAttributes();
	}
	
	/*
	 * Start a meta element.
	 */
	private void startMetaElement(Attributes attributes) {
		if (factory == null || attributeList != null) {
			// Either not in xesextension element or in log, other trace, event, or meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, XesXmlTags.META, XesXmlTags.XESEXT);
			issue.setLocation(locator);
			issues.add(issue);
		} else if (!extension.getMetaAttributes().isEmpty()) {
			// Meta element may not succeed another meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueOutOfPlace(extension, XesXmlTags.META, XesXmlTags.META);
			issue.setLocation(locator);
			issues.add(issue);
		}

		for (int i = 0; i < attributes.getLength(); i++) {
			String value = attributes.getQName(i);
			XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value, "Clause 6.5.2");
			issue.setLocation(locator);
			issues.add(issue);
		}
		
		// Set the list for the child attributes.
		attributeList = extension.getMetaAttributes();
	}
	
	/*
	 * Start a string element.
	 */
	private void startStringElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createStringAttribute(), XesXmlTags.STRING);
	}
	
	/*
	 * Start a date element.
	 */
	private void startDateElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createDateTimeAttribute(), XesXmlTags.DATE);
	}
	
	/*
	 * Start an int element.
	 */
	private void startIntegerNumberElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createIntegerNumberAttribute(), XesXmlTags.INT);
	}
	
	/*
	 * Start a float element.
	 */
	private void startRealNumberElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createRealNumberAttribute(), XesXmlTags.FLOAT);
	}
	
	/*
	 * Start a boolean element.
	 */
	private void startBooleanElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createBooleanAttribute(), XesXmlTags.BOOLEAN);
	}
	
	/*
	 * Start an id element.
	 */
	private void startIdElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createIdAttribute(), XesXmlTags.ID);
	}
	
	/*
	 * Start a list element.
	 */
	private void startListElement(Attributes attributes) {
		startAttributeElement(attributes, factory.createListAttribute(), XesXmlTags.LIST);
	}
	
	/*
	 * Start an attribute element. The provided attribute should be of the correct type, and
	 * the provided label should match the type.
	 */
	private void startAttributeElement(Attributes attributes, XesAttribute attribute, String label) {
		if (attributeList == null) {
			// No in any attribute list, hence not in a log, trace, event, or meta element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, label,
					XesXmlTags.LOG + ", " + XesXmlTags.TRACE + ", " + XesXmlTags.EVENT + ", or " + XesXmlTags.META);
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			
			// Check required key attribute.
			String keyValue = attributes.getValue(XesXmlTags.KEY);
			if (keyValue == null) {
				// No key attribute. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, label,
						XesXmlTags.KEY, "Clause 6.6.2");
				issue.setLocation(locator);
				issues.add(issue);
			} else {
				// Set the prefixed attribute key.
				attribute.setKey(extension.getPrefix() + ":" + keyValue);
				attributeList.add(attribute);
				// Se tthe current attribute (for any alias child elements).
				currentAttribute = attribute;
			}

			// Check for additional attributes.
			for (int i = 0; i < attributes.getLength(); i++) {
				String value = attributes.getQName(i);
				if (value.equals(XesXmlTags.KEY)) {
				} else {
					// Additional attribute found, Raise issue.
					XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value,
							"Clause 6.6.2");
					issue.setLocation(locator);
					issues.add(issue);
				}
			}
		}
	}
	
	/*
	 * Start an alias element.
	 */
	private void startAliasElement(Attributes attributes) {
		
		if (currentAttribute == null) {
			// No current attribute, hence not in attribute element. Raise issue.
			XesXmlIssue issue = new XesextXmlIssueParent(extension, XesXmlTags.ALIAS, "attribute");
			issue.setLocation(locator);
			issues.add(issue);
		} else {
			// Check required mapping attribute.
			String mappingValue = attributes.getValue(XesXmlTags.MAPPING);
			if (mappingValue == null) {
				// No valid mapping attribute. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, XesXmlTags.ALIAS,
						XesXmlTags.MAPPING, "Clause 6.7.2");
				issue.setLocation(locator);
				issues.add(issue);
				mappingValue = "<no mapping>";
			} else if (!ISO_LANGUAGES.contains(mappingValue.toLowerCase())){
				// MappingValue does not correspond to an ISO language code. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueLanguageCode(extension, mappingValue);
				issue.setLocation(locator);
				issues.add(issue);
			}
			
			// Check required name attribute. 
			String nameValue = attributes.getValue(XesXmlTags.NAME);
			if (nameValue == null) {
				// No name attribute. Raise issue.
				XesXmlIssue issue = new XesextXmlIssueAttributeRequired(extension, XesXmlTags.ALIAS,
						XesXmlTags.NAME, "Clause 6.7.2");
				issue.setLocation(locator);
				issues.add(issue);
				nameValue = "<no name>";
			}

			// Check for additional attributes.
			for (int i = 0; i < attributes.getLength(); i++) {
				String value = attributes.getQName(i);
				if (value.equals(XesXmlTags.MAPPING) || value.equals(XesXmlTags.NAME)) {
				} else {
					// Found additional attribute. Raise issue.
					XesXmlIssue issue = new XesextXmlIssueStrictlyConformingAttribute(extension, value,
							"Clause 6.7.2");
					issue.setLocation(locator);
					issues.add(issue);
				}
			}

			// Add the alias to the current attribute.
			extension.addAlias(currentAttribute, mappingValue, nameValue);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String,
	 * java.lang.String, java.lang.String)
	 */
	public void endElement(String uri, String localName, String qName) throws SAXException {
		String tagName = localName;
		if (tagName.equalsIgnoreCase("")) {
			tagName = qName;
		}

		if (tagName.equalsIgnoreCase(XesXmlTags.XESEXT)) {
			factory = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.LOG)) {
			attributeList = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.TRACE)) {
			attributeList = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.EVENT)) {
			attributeList = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.META)) {
			attributeList = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.STRING)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.DATE)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.INT)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.FLOAT)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.BOOLEAN)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.ID)) {
			currentAttribute = null;

		} else if (tagName.equalsIgnoreCase(XesXmlTags.LIST)) {
			currentAttribute = null;

		}
	}
}
